#ifndef NPSTAT_ABSMARGINALSMOOTHER_HH_
#define NPSTAT_ABSMARGINALSMOOTHER_HH_

/*!
// \file AbsMarginalSmoother.hh
//
// \brief Interface definition for 1-d nonparametric density estimation
//
// Author: I. Volobouev
//
// September 2010
*/

#include "npstat/stat/AbsPolyFilter1D.hh"
#include "npstat/stat/HistoND.hh"

namespace npstat {
    /**
    // Interface definition for classes which build gridded 1-d distrubutions
    // out of data samples (typically, by LOrPE, KDE, or some other
    // nonparametric techniques). The derived classes are supposed to
    // override the "smoothHistogram" member function as well as the
    // "selfContribution" function inherited from the AbsPolyFilter1D base.
    */
    class AbsMarginalSmoother : public AbsPolyFilter1D
    {
    public:
        /** Constructor takes binning parameters for the internal histogram */
        AbsMarginalSmoother(unsigned nbins, double xmin, double xmax);

        virtual ~AbsMarginalSmoother() {}

        //@{
        /** Simple inspector of the object properties */
        inline unsigned dataLen() const {return histo_.nBins();}
        inline double xMin() const {return histo_.axis(0U).min();}
        inline double xMax() const {return histo_.axis(0U).max();}
        inline double binWidth() const {return histo_.binVolume();}
        //@}

        /**
        // The "bandwidthUsed" pointer in this method can be NULL.
        // The method returns a reference to an internal histogram which
        // will remain unchanged until the next call to "smooth" or until
        // this object is destroyed. "inputPoints" is a collection of
        // scattered 1-d data points which will be histogrammed first
        // and then smoothed.
        */
        template <typename Numeric>
        const HistoND<double>& smooth(const std::vector<Numeric>& inputPoints,
                                      double* bandwidthUsed = 0);
    private:
        AbsMarginalSmoother();
        AbsMarginalSmoother(const AbsMarginalSmoother&);
        AbsMarginalSmoother& operator=(const AbsMarginalSmoother&);

        /**
        // The "bandwidthUsed" pointer in this method can be NULL.
        // This is the function to implement in the derived classes.
        */
        virtual void smoothHistogram(HistoND<double>& histo,
                                     double* bandwidthUsed) = 0;
        HistoND<double> histo_;
    };
}

#include "npstat/stat/AbsMarginalSmoother.icc"

#endif // NPSTAT_ABSMARGINALSMOOTHER_HH_
