#ifndef NPSTAT_CVCOPULASMOOTHER_HH_
#define NPSTAT_CVCOPULASMOOTHER_HH_

/*!
// \file CVCopulaSmoother.hh
//
// \brief Interface definition for smoothing copulas with LOrPE
//        and cross-validation
//
// Abbreviation "CV" used in various places in this code
// stands for "cross-validation".
//
// Author: I. Volobouev
//
// September 2010
*/

#include "npstat/stat/AbsCopulaSmoother.hh"

namespace npstat {
    /**
    // An interface for copula smoothers which use constant bandwidth
    // LOrPE. A cross validation algorithm (derived from AbsBandwidthCVND)
    // is employed to choose the bandwidth. The derived classes are supposed
    // to build LOrPE filters in their constructors and then call "setFilter"
    // method of this class for each filter constructed.
    //
    // Note that the constructor of this class is protected, so there is
    // no way to build an instance of this class directly.
    */
    template<class F>
    class CVCopulaSmoother : public AbsCopulaSmoother
    {
    public:
        typedef AbsBandwidthCVND<double,ArrayND<double> > CVCalc;
        typedef F Filter;

        virtual ~CVCopulaSmoother();

        inline bool isConvoluting() const {return useConvolute_;}
        inline void setConvoluting(const bool b) {useConvolute_ = b;}

    protected:
        /**
        // Constructor arguments are as follows:
        //
        // nBinsInEachDim    -- number of copula bins in each dimension
        //
        // dim               -- copula dimensionality
        //
        // marginTolerance   -- tolerance for the margin to be uniform
        //
        // maxNormCycles     -- max number of copula normalization cycles
        //
        // initialBw         -- "central" bandwidth for cross validation
        //                      calculations (or the actual bandwidth used
        //                      in case cross validation is not performed).
        //                      Set this parameter to 0.0 in order to
        //                      disable filtering altogether.
        //
        // cvCalc            -- calculator for the quantity being optimized
        //                      in the cross validation process. May be NULL
        //                      in which case cross validation will not
        //                      be used.
        //
        // becomeCvCalcOwner -- tells us whether we should destroy cvCalc
        //                      in our own destructor
        //
        // cvRange           -- we will scan bandwidth values between
        //                      initialBw/cvRange and initialBw*cvRange
        //                      uniformly in the log space.
        //
        // nCV               -- number of bandwidth values to try in the
        //                      bandwidth scan. If this number is even, it
        //                      will be increased by 1 internally so that
        //                      the "central" bandwidth is included in
        //                      the scan. If this parameter is 0 or 1, the
        //                      value given by "initialBw" will be used.
        //
        // useConvolve       -- if "true", use "convolve" method of the
        //                      filter rather than "filter" method.
        */
        CVCopulaSmoother(const unsigned* nBinsInEachDim, unsigned dim,
                         double marginTolerance,
                         unsigned maxNormCycles, double initialBw,
                         const CVCalc* cvCalc, bool becomeCvCalcOwner,
                         double cvRange, unsigned nCV, bool useConvolve);

        inline const std::vector<double>& bandwidthValues() const
            {return bandwidthValues_;}

        /**
        // Constructors of the derived classes should call
        // this method for each bandwidth value. "i" is
        // the bandwidth value number, while the bandwidth
        // corresponding to this number should be obtained from
        // the "bandwidthValues()" vector.
        //
        // This base class will assume the ownership of the filter object.
        */
        void setFilter(unsigned i, Filter* filter);

    private:
        CVCopulaSmoother();
        CVCopulaSmoother(const CVCopulaSmoother&);
        CVCopulaSmoother& operator=(const CVCopulaSmoother&);

        void smoothHistogram(HistoND<double>& histo, double* bandwidthUsed);

        // The following vectors must have the same size
        std::vector<Filter*> filters_;
        std::vector<double> bandwidthValues_;

        ArrayND<double> density_;
        const CVCalc* cvCalc_;
        bool assumeCvCalcOwnership_;
        bool doFilter_;
        bool useConvolute_;
    };
}

#include "npstat/stat/CVCopulaSmoother.icc"

#endif // NPSTAT_CVCOPULASMOOTHER_HH_
