#include <cmath>
#include <cassert>
#include <stdexcept>
#include <climits>
#include <cfloat>
#include <algorithm>

#include "npstat/nm/OrthoPoly1D.hh"
#include "npstat/nm/findRootInLogSpace.hh"

#include "npstat/stat/Filter1DBuilders.hh"

static const double doubleOne = 1.0;

static npstat::OrthoPoly1D* build(
    unsigned* filterCenter, const unsigned maxDegree,
    const double* weight, const unsigned len,
    const unsigned center, const double step)
{
    *filterCenter = center;
    return new npstat::OrthoPoly1D(maxDegree, weight, len, step);
}

static inline int mirrorIndex(const int i, const int period)
{
    int j = i % period;
    if (j < 0)
        j += period;
    int rat = i/period;
    if (i < 0 && i % period)
        --rat;
    if (rat % 2)
        return period - j - 1;
    else
        return j;
}

namespace npstat {
    WeightTableFilter1DBuilder::WeightTableFilter1DBuilder(
        const double* weight, const unsigned weightLen,
        const unsigned char* exclusionMask, const unsigned exclusionMaskLen,
        const bool excludeCentralPoint)
    {
        if (!weightLen) throw std::invalid_argument(
            "In npstat::WeightTableFilter1DBuilder constructor: "
            "array of weights must not be empty");
        assert(weight);

        // Construct the full weight
        w_.resize(2*weightLen - 1);
        w_[weightLen - 1] = excludeCentralPoint ? 0.0 : weight[0];       
        for (unsigned i=1; i<weightLen; ++i)
        {
            w_[weightLen + i - 1] = weight[i];
            w_[weightLen - i - 1] = weight[i];
        }

        // Fill the exclusion mask
        if (exclusionMask && exclusionMaskLen)
        {
            exclusionMask_.resize(exclusionMaskLen);
            for (unsigned i=0; i<exclusionMaskLen; ++i)
                exclusionMask_[i] = exclusionMask[i];
            wexcl_.resize(w_.size());
        }
    }

    OrthoPoly1D* WeightTableFilter1DBuilder::makeOrthoPoly(
        const unsigned maxDeg, const unsigned i,
        const unsigned dataLen, unsigned* filterCenter) const
    {
        if (!dataLen) throw std::invalid_argument(
            "In npstat::WeightTableFilter1DBuilder::makeOrthoPoly: "
            "invalid data length parameter");
        assert(filterCenter);
        assert(i < dataLen);

        const double step = 1.0/sqrt(1.0*w_.size());
        const unsigned weightLen = w_.size();
        const unsigned n_half = weightLen/2;

        const bool hasExcl = !exclusionMask_.empty();
        if (hasExcl)
        {
            if (exclusionMask_.size() != dataLen)
                throw std::invalid_argument(
                    "In npstat::WeightTableFilter1DBuilder::makeOrthoPoly: "
                    "data length is not compatible with the exclusion mask");

            // Weight element number n_half corresponds to mask element i
            for (unsigned iw=0; iw<weightLen; ++iw)
            {
                wexcl_[iw] = w_[iw];
                const int imask = static_cast<int>(iw + i) -
                                  static_cast<int>(n_half);
                if (imask >= 0 && imask < static_cast<int>(dataLen))
                    if (exclusionMask_[imask])
                        wexcl_[iw] = 0.0;
            }
        }

        if (i < n_half)
        {
            // Create left boundary weight and polynomials
            unsigned n_use = n_half + i + 1;
            if (n_use > dataLen)
                n_use = dataLen;
            const double* wptr = hasExcl ? &wexcl_[n_half-i] : &w_[n_half-i];
            return build(filterCenter, maxDeg, wptr, n_use, i, step);
        }
        else if (i + n_half >= dataLen)
        {
            // Create right boundary weight and polynomials
            const unsigned n_use = dataLen + n_half - i;
            assert(n_use <= dataLen);
            const double* wptr = hasExcl ? &wexcl_[0] : &w_[0];
            return build(filterCenter, maxDeg, wptr, n_use, n_half, step);
        }
        else
        {
            // Symmetric weight
            const double* wptr = hasExcl ? &wexcl_[0] : &w_[0];
            return build(filterCenter, maxDeg, wptr, w_.size(), n_half, step);
        }
    }

    StretchingFilter1DBuilder::StretchingFilter1DBuilder(
        const AbsDistribution1D* distro, const double step,
        const unsigned char* exclusionMask, const unsigned exclusionMaskLen,
        const bool excludeCentralPoint)
        : distro_(distro),
          step_(step),
          excludeCentralPoint_(excludeCentralPoint)
    {
        if (step_ <= 0.0) throw std::invalid_argument(
            "In npstat::StretchingFilter1DBuilder constructor: "
            "invalid step size");
        assert(distro_);

        const double maxratio = floor(fabs(distro_->quantile(1.0))/step_);
        const double minratio = floor(fabs(distro_->quantile(0.0))/step_);
        const double topratio = maxratio > minratio ? maxratio : minratio;
        if (topratio >= static_cast<double>(INT_MAX))
            maxlen_ = 2*(static_cast<unsigned>(INT_MAX) - 1) + 1;
        else
            maxlen_ = 2*static_cast<unsigned>(topratio) + 1;

        // Fill the exclusion mask
        if (exclusionMask && exclusionMaskLen)
        {
            exclusionMask_.resize(exclusionMaskLen);
            for (unsigned i=0; i<exclusionMaskLen; ++i)
                exclusionMask_[i] = exclusionMask[i];
        }
    }

    OrthoPoly1D* StretchingFilter1DBuilder::makeOrthoPoly(
        const unsigned maxDeg, const unsigned i,
        const unsigned dataLen, unsigned* filterCenter) const
    {
        if (!(dataLen && dataLen <= static_cast<unsigned>(INT_MAX)))
            throw std::invalid_argument(
                "In npstat::StretchingFilter1DBuilder::makeOrthoPoly: "
                "invalid data length parameter");
        assert(filterCenter);
        assert(i < dataLen);

        const bool hasExcl = !exclusionMask_.empty();
        if (hasExcl)
            if (exclusionMask_.size() != dataLen)
                throw std::invalid_argument(
                    "In npstat::StretchingFilter1DBuilder::makeOrthoPoly: "
                    "data length is not compatible with the exclusion mask");

        // Make sure we have enough space
        const unsigned sizeNeeded = (dataLen < maxlen_ ? dataLen : maxlen_);
        if (w_.size() < sizeNeeded)
            w_.resize(sizeNeeded);

        // Figure out the maximum possible efficiency of the weight scan
        const double maxeff = distro_->cdf((dataLen - 1 - dataLen/2)*step_) - 
                              distro_->cdf(-step_*(dataLen/2));
        if (maxeff <= 0.0) throw std::invalid_argument(
            "In npstat::StretchingFilter1DBuilder::makeOrthoPoly: "
            "weight density is identically zero in the relevant region");

        // Figure out the efficiency for this bin
        const double eff = distro_->cdf((dataLen - 1 - i)*step_) - 
                           distro_->cdf(-step_*i);

        // Stretch the function so that it covers wider range if
        // the efficiency is small. Stretching the function is
        // equivalent to reducing the step size used for scanning
        // the function.
        const double scanStep = step_*eff/maxeff;

        // Figure out where the center bin is going to be
        int j0;
        const unsigned n_half = maxlen_/2;
        if (i < n_half)
        {
            // Create the left boundary weight and polynomials.
            // The center is going to be at i.
            j0 = static_cast<int>(i);
        }
        else if (i + n_half >= dataLen)
        {
            // Create the right boundary weight and polynomials.
            // The center is going to be at sizeNeeded - (dataLen - i).
            assert(sizeNeeded >= dataLen - i);
            j0 = static_cast<int>(sizeNeeded - (dataLen - i));
        }
        else
        {
            // Symmetric support region
            assert(maxlen_ <= dataLen);
            j0 = static_cast<int>(n_half);
        }

        const int jmax = static_cast<int>(sizeNeeded);
        assert(j0 >= 0 && j0 < jmax);
        for (int j=0; j<jmax; ++j)
        {
            w_[j] = distro_->density(scanStep*(j - j0));

            // Process the exclusions
            if (hasExcl)
            {
                // bin j0 of the weight corresponds to the mask bin i
                const int imask = j + static_cast<int>(i) - j0;
                if (imask >= 0 && imask < static_cast<int>(dataLen))
                    if (exclusionMask_[imask])
                        w_[j] = 0.0;
            }
        }

        if (excludeCentralPoint_)
            w_[j0] = 0.0;

        *filterCenter = j0;
        return new npstat::OrthoPoly1D(maxDeg, &w_[0], sizeNeeded, step_);
    }

    FoldingFilter1DBuilder::FoldingFilter1DBuilder(
        const AbsDistribution1D* distro, const double step,
        const unsigned char* exclusionMask, const unsigned exclusionMaskLen,
        const bool excludeCentralPoint)
        : distro_(distro),
          step_(step),
          excludeCentralPoint_(excludeCentralPoint)
    {
        if (step_ <= 0.0) throw std::invalid_argument(
            "In npstat::FoldingFilter1DBuilder constructor: invalid step size");
        assert(distro_);

        const double maxratio = floor(fabs(distro_->quantile(1.0))/step_);
        const double minratio = floor(fabs(distro_->quantile(0.0))/step_);
        const double topratio = maxratio > minratio ? maxratio : minratio;
        if (topratio >= static_cast<double>(INT_MAX))
            maxlen_ = 2*(static_cast<unsigned>(INT_MAX) - 1) + 1;
        else
            maxlen_ = 2*static_cast<unsigned>(topratio) + 1;

        // Fill the exclusion mask
        if (exclusionMask && exclusionMaskLen)
        {
            exclusionMask_.resize(exclusionMaskLen);
            for (unsigned i=0; i<exclusionMaskLen; ++i)
                exclusionMask_[i] = exclusionMask[i];
        }
    }

    OrthoPoly1D* FoldingFilter1DBuilder::makeOrthoPoly(
        const unsigned maxDeg, const unsigned i,
        const unsigned dataLen, unsigned* filterCenter) const
    {
        if (!(dataLen && dataLen <= static_cast<unsigned>(INT_MAX)))
            throw std::invalid_argument(
                "In npstat::FoldingFilter1DBuilder::makeOrthoPoly: "
                "invalid data length parameter");
        assert(filterCenter);
        assert(i < dataLen);

        const bool hasExcl = !exclusionMask_.empty();
        if (hasExcl)
            if (exclusionMask_.size() != dataLen)
                throw std::invalid_argument(
                    "In npstat::FoldingFilter1DBuilder::makeOrthoPoly: "
                    "data length is not compatible with the exclusion mask");

        // Make sure we have enough space
        const unsigned sizeNeeded = (dataLen < maxlen_ ? dataLen : maxlen_);
        if (w_.size() < sizeNeeded)
            w_.resize(sizeNeeded);

        // Figure out where the center bin is going to be
        int j0;
        bool isBoundaryKernel = true;
        int patchOffset = 0;
        const unsigned n_half = maxlen_/2;
        if (i < n_half)
        {
            // Create the left boundary weight and polynomials.
            // The center is going to be at i.
            j0 = static_cast<int>(i);
        }
        else if (i + n_half >= dataLen)
        {
            // Create the right boundary weight and polynomials.
            // The center is going to be at sizeNeeded - (dataLen - i).
            assert(sizeNeeded >= dataLen - i);
            j0 = static_cast<int>(sizeNeeded - (dataLen - i));
            patchOffset = dataLen - sizeNeeded;
        }
        else
        {
            // Symmetric support region
            assert(maxlen_ <= dataLen);
            j0 = static_cast<int>(n_half);
            isBoundaryKernel = false;
        }

        const int jmax = static_cast<int>(sizeNeeded);

        if (isBoundaryKernel)
        {
            for (unsigned k=0; k<sizeNeeded; ++k)
                w_[k] = 0.0;
            w_[j0] = distro_->density(0.0);
            const int halfn = n_half;
            for (int k=1; k<=halfn; ++k)
            {
                for (unsigned idir=0; idir<2; ++idir)
                {
                    const int j = idir ? k : -k;
                    const int mi = mirrorIndex(patchOffset + j0 + j, dataLen);
                    const int indexInPatch = mi - patchOffset;
                    assert(indexInPatch >= 0 && indexInPatch < jmax);
                    w_[indexInPatch] += distro_->density(step_*j);
                }
            }
        }
        else
        {
            assert(j0 >= 0 && j0 < jmax);
            for (int j=0; j<jmax; ++j)
                w_[j] = distro_->density(step_*(j - j0));
        }

        // Process the exclusions
        if (hasExcl)
            for (int j=0; j<jmax; ++j)
            {
                // bin j0 of the weight corresponds to the mask bin i
                const int imask = j + static_cast<int>(i) - j0;
                if (imask >= 0 && imask < static_cast<int>(dataLen))
                    if (exclusionMask_[imask])
                        w_[j] = 0.0;
            }

        if (excludeCentralPoint_)
            w_[j0] = 0.0;

        *filterCenter = j0;
        return new npstat::OrthoPoly1D(maxDeg, &w_[0], sizeNeeded, step_);
    }

    FoldingSqFilter1DBuilder::FoldingSqFilter1DBuilder(
        const AbsDistribution1D* distro, const double step,
        const unsigned char* exclusionMask, const unsigned exclusionMaskLen,
        const bool excludeCentralPoint)
        : distro_(distro),
          step_(step),
          excludeCentralPoint_(excludeCentralPoint)
    {
        if (step_ <= 0.0) throw std::invalid_argument(
            "In npstat::FoldingSqFilter1DBuilder constructor: invalid step size");
        assert(distro_);

        // Fill the exclusion mask
        if (exclusionMask && exclusionMaskLen)
        {
            exclusionMask_.resize(exclusionMaskLen);
            for (unsigned i=0; i<exclusionMaskLen; ++i)
                exclusionMask_[i] = exclusionMask[i];
        }

        // Calculate the integral of the normalized squared weight function
        long double fsum = 0.0L, sum = 0.0L;
        maxlen_ = INT_MAX - 1;
        int lolen = -static_cast<int>(maxlen_);
        for (int i=0; i<INT_MAX; ++i)
        {
            const double x = i*step_;
            const double dens = distro_->density(x);
            if (dens <= 0.0)
            {
                maxlen_ = i;
                break;
            }
            sum += dens*dens;
            fsum += dens;
        }
        for (int i=-1; i>INT_MIN; --i)
        {
            const double x = i*step_;
            const double dens = distro_->density(x);
            if (dens <= 0.0)
            {
                lolen = i;
                break;
            }
            sum += dens*dens;
            fsum += dens;
        }
        assert(fsum > 0.0L);
        const double fscale = 1.0/(fsum*step_);
        integ_ = sum*step_*fscale*fscale;
        assert(integ_ > 0.0);

        const unsigned utmp = abs(lolen);
        if (utmp > maxlen_)
            maxlen_ = utmp;
        maxlen_ = 2*(maxlen_ - 1U) + 1;
    }

    double FoldingSqFilter1DBuilder::integrateSquaredWeight(
        const unsigned datalen, const unsigned binnum, const double h) const
    {
        assert(binnum < datalen);
        assert(h > 0.0);

        if (w_.size() < datalen)
            w_.resize(datalen);
        double* w = &w_[0];
        {
            double* wpt = w;
            for (unsigned k=0; k<datalen; ++k)
                *wpt++ = 0.0;
        }
        w[binnum] = distro_->density(0.0)/h;

        const int len = datalen;
        const int j0 = binnum;
        const int kmax = INT_MAX-len;
        for (int k=1; k<kmax; ++k)
        {
            const double x = step_*k;
            const double dens = distro_->density(x/h)/h;
            if (dens <= 0.0)
                break;
            w[mirrorIndex(j0 + k, len)] += dens;
        }
        const int kmin = INT_MIN+len+1;
        for (int k=-1; k>kmin; --k)
        {
            const double x = step_*k;
            const double dens = distro_->density(x/h)/h;
            if (dens <= 0.0)
                break;
            w[mirrorIndex(j0 + k, len)] += dens;
        }

        long double fsum = 0.0L, sum = 0.0L;
        for (unsigned i=0; i<datalen; ++i)
        {
            fsum += w[i];
            sum += w[i]*w[i];
        }
        assert(fsum > 0.0L);
        const double fscale = 1.0/(fsum*step_);
        return sum*step_*fscale*fscale;
    }

    OrthoPoly1D* FoldingSqFilter1DBuilder::makeOrthoPoly(
        const unsigned maxDeg, const unsigned i,
        const unsigned dataLen, unsigned* filterCenter) const
    {
        if (!(dataLen && dataLen <= static_cast<unsigned>(INT_MAX)))
            throw std::invalid_argument(
                "In npstat::FoldingSqFilter1DBuilder::makeOrthoPoly: "
                "invalid data length parameter");
        assert(filterCenter);
        assert(i < dataLen);

        const bool hasExcl = !exclusionMask_.empty();
        if (hasExcl)
            if (exclusionMask_.size() != dataLen)
                throw std::invalid_argument(
                    "In npstat::FoldingSqFilter1DBuilder::makeOrthoPoly: "
                    "data length is not compatible with the exclusion mask");

        // Make sure we have enough space
        if (w_.size() < dataLen)
            w_.resize(dataLen);
        double* wfirst = &w_[0];

        // Figure out where the center bin is going to be
        const unsigned n_half = maxlen_/2;
        const bool isBoundaryKernel = i < n_half || i + n_half >= dataLen;

        int j0;
        unsigned sizeNeeded;

        if (isBoundaryKernel)
        {
            double maxinteg = integ_;
            if (dataLen < maxlen_)
                // Need to recalculate max possible value
                // of weight integral squared
                maxinteg = integrateSquaredWeight(dataLen, dataLen/2, 1.0);

            double h = 1.0;
            double currentInteg = integrateSquaredWeight(dataLen, i, h);
            if (currentInteg > maxinteg)
            {
                // h is expected to be in the [1, 2] interval
                const double tol = 2.0*DBL_EPSILON*sqrt(dataLen) + 1.0e-14;
                const bool status = findRootInLogSpace(
                    Fnc(*this, dataLen, i), maxinteg,
                    sqrt(2.0), tol, &h, 0.5*log(2.01/0.99));
                assert(status);
                integrateSquaredWeight(dataLen, i, h);
            }

            // Redo j0, wfirst, and sizeNeeded
            wfirst = &w_[0];
            j0 = i;
            sizeNeeded = dataLen;
            for (unsigned i=0; i<dataLen; ++i)
            {
                if (*wfirst == 0.0)
                {
                    wfirst++;
                    j0--;
                    sizeNeeded--;
                }
                else
                    break;
            }
            for (unsigned i=dataLen-1; i>0; --i)
            {
                if (w_[i] == 0.0)
                    sizeNeeded--;
                else
                    break;
            }
        }
        else
        {
            assert(dataLen >= maxlen_);
            j0 = static_cast<int>(n_half);
            sizeNeeded = maxlen_;
            const int jmax = sizeNeeded;
            for (int j=0; j<jmax; ++j)
                wfirst[j] = distro_->density(step_*(j - j0));
        }

        // Process the exclusions
        if (hasExcl)
        {
            const int jmax = sizeNeeded;
            for (int j=0; j<jmax; ++j)
            {
                // bin j0 of the weight corresponds to the mask bin i
                const int imask = j + static_cast<int>(i) - j0;
                if (imask >= 0 && imask < static_cast<int>(dataLen))
                    if (exclusionMask_[imask])
                        wfirst[j] = 0.0;
            }
        }

        if (excludeCentralPoint_)
            wfirst[j0] = 0.0;

        *filterCenter = j0;
        return new npstat::OrthoPoly1D(maxDeg, wfirst, sizeNeeded, step_);
    }

    NonmodifyingFilter1DBuilder::NonmodifyingFilter1DBuilder()
        : WeightTableFilter1DBuilder(&doubleOne, 1U)
    {
    }

    OrthoPoly1D* NonmodifyingFilter1DBuilder::makeOrthoPoly(
        unsigned /* maxDegree */, const unsigned binnum,
        const unsigned datalen, unsigned* filterCenter) const
    {
        return WeightTableFilter1DBuilder::makeOrthoPoly(
            0U, binnum, datalen, filterCenter);
    }

    ConstSqFilter1DBuilder::ConstSqFilter1DBuilder(
        const AbsDistribution1D* distro, const double step,
        const unsigned char* exclusionMask, const unsigned exclusionMaskLen,
        const bool excludeCentralPoint)
        : distro_(distro),
          step_(step),
          excludeCentralPoint_(excludeCentralPoint)
    {
        if (step_ <= 0.0) throw std::invalid_argument(
            "In npstat::ConstSqFilter1DBuilder constructor: invalid step size");
        assert(distro_);

        // Fill the exclusion mask
        if (exclusionMask && exclusionMaskLen)
        {
            exclusionMask_.resize(exclusionMaskLen);
            for (unsigned i=0; i<exclusionMaskLen; ++i)
                exclusionMask_[i] = exclusionMask[i];
        }

        // Calculate the integral of the normalized squared weight function
        long double fsum = 0.0L, sum = 0.0L;
        maxlen_ = INT_MAX - 1;
        int lolen = -static_cast<int>(maxlen_);
        for (int i=0; i<INT_MAX; ++i)
        {
            const double x = i*step_;
            const double dens = distro_->density(x);
            if (dens <= 0.0)
            {
                maxlen_ = i;
                break;
            }
            sum += dens*dens;
            fsum += dens;
        }
        for (int i=-1; i>INT_MIN; --i)
        {
            const double x = i*step_;
            const double dens = distro_->density(x);
            if (dens <= 0.0)
            {
                lolen = i;
                break;
            }
            sum += dens*dens;
            fsum += dens;
        }
        assert(fsum > 0.0L);
        const double fscale = 1.0/(fsum*step_);
        integ_ = sum*step_*fscale*fscale;
        assert(integ_ > 0.0);

        const unsigned utmp = abs(lolen);
        if (utmp > maxlen_)
            maxlen_ = utmp;
        maxlen_ = 2*(maxlen_ - 1U) + 1;
    }

    double ConstSqFilter1DBuilder::integrateSquaredWeight(
        const unsigned datalen, const unsigned binnum, const double h) const
    {
        assert(binnum < datalen);
        assert(h > 0.0);

        long double fsum = 0.0L, sum = 0.0L;
        for (unsigned i=binnum; i<datalen; ++i)
        {
            const double x = (i-binnum)*step_;
            const double dens = distro_->density(x/h)/h;
            if (dens <= 0.0)
                break;
            fsum += dens;
            sum += dens*dens;
        }
        for (unsigned i=binnum; i>0U; --i)
        {
            const double x = -step_*(binnum-i+1U);
            const double dens = distro_->density(x/h)/h;
            if (dens <= 0.0)
                break;
            fsum += dens;
            sum += dens*dens;
        }
        assert(fsum > 0.0L);
        const double fscale = 1.0/(fsum*step_);
        return sum*step_*fscale*fscale;
    }

    OrthoPoly1D* ConstSqFilter1DBuilder::makeOrthoPoly(
        const unsigned maxDeg, const unsigned i,
        const unsigned dataLen, unsigned* filterCenter) const
    {
        if (!(dataLen && dataLen <= static_cast<unsigned>(INT_MAX)))
            throw std::invalid_argument(
                "In npstat::ConstSqFilter1DBuilder::makeOrthoPoly: "
                "invalid data length parameter");
        assert(filterCenter);
        assert(i < dataLen);

        const bool hasExcl = !exclusionMask_.empty();
        if (hasExcl)
            if (exclusionMask_.size() != dataLen)
                throw std::invalid_argument(
                    "In npstat::ConstSqFilter1DBuilder::makeOrthoPoly: "
                    "data length is not compatible with the exclusion mask");

        // Make sure we have enough space
        const unsigned sizeNeeded = (dataLen < maxlen_ ? dataLen : maxlen_);
        if (w_.size() < sizeNeeded)
            w_.resize(sizeNeeded);

        // Figure out the maximum possible efficiency of the weight scan
        const double maxeff = distro_->cdf((dataLen - 1 - dataLen/2)*step_) - 
                              distro_->cdf(-step_*(dataLen/2));
        if (maxeff <= 0.0) throw std::invalid_argument(
            "In npstat::ConstSqFilter1DBuilder::makeOrthoPoly: "
            "weight density is identically zero in the relevant region");

        // Figure out the efficiency for this bin
        const double eff = distro_->cdf((dataLen - 1 - i)*step_) - 
                           distro_->cdf(-step_*i);

        // Stretching the function is equivalent to reducing
        // the step size used for scanning the function
        double scanStep = step_;
        if (eff/maxeff < 1.0 - 8.0*DBL_EPSILON)
        {
            double maxinteg = integ_;
            if (dataLen < maxlen_)
                // Need to recalculate max possible value
                // of weight integral squared
                maxinteg = integrateSquaredWeight(dataLen, dataLen/2, 1.0);

            double h = 1.0;
            double currentInteg = integrateSquaredWeight(dataLen, i, h);
            if (currentInteg > maxinteg)
            {
                // h is expected to be in the [1, 2] interval
                const double tol = 2.0*DBL_EPSILON*sqrt(dataLen) + 1.0e-14;
                const bool status = findRootInLogSpace(
                    Fnc(*this, dataLen, i), maxinteg,
                    sqrt(2.0), tol, &h, 0.5*log(2.01/0.99));
                assert(status);
            }
            scanStep = step_/h;
        }

        // Figure out where the center bin is going to be
        int j0;
        const unsigned n_half = maxlen_/2;
        if (i < n_half)
        {
            // Create the left boundary weight and polynomials.
            // The center is going to be at i.
            j0 = static_cast<int>(i);
        }
        else if (i + n_half >= dataLen)
        {
            // Create the right boundary weight and polynomials.
            // The center is going to be at sizeNeeded - (dataLen - i).
            assert(sizeNeeded >= dataLen - i);
            j0 = static_cast<int>(sizeNeeded - (dataLen - i));
        }
        else
        {
            // Symmetric support region
            assert(maxlen_ <= dataLen);
            j0 = static_cast<int>(n_half);
        }

        const int jmax = static_cast<int>(sizeNeeded);
        assert(j0 >= 0 && j0 < jmax);
        for (int j=0; j<jmax; ++j)
        {
            w_[j] = distro_->density(scanStep*(j - j0));

            // Process the exclusions
            if (hasExcl)
            {
                // bin j0 of the weight corresponds to the mask bin i
                const int imask = j + static_cast<int>(i) - j0;
                if (imask >= 0 && imask < static_cast<int>(dataLen))
                    if (exclusionMask_[imask])
                        w_[j] = 0.0;
            }
        }

        if (excludeCentralPoint_)
            w_[j0] = 0.0;

        *filterCenter = j0;
        return new npstat::OrthoPoly1D(maxDeg, &w_[0], sizeNeeded, step_);
    }
}
