#ifndef NPSTAT_FILTER1DBUILDERS_HH_
#define NPSTAT_FILTER1DBUILDERS_HH_

/*!
// \file Filter1DBuilders.hh
//
// \brief Builders of local polynomial filters in one dimension
//
// The filters are intended for use with the "LocalPolyFilter1D" class.
//
// Author: I. Volobouev
//
// November 2009
*/

#include <vector>

#include "npstat/nm/SimpleFunctors.hh"

#include "npstat/stat/AbsFilter1DBuilder.hh"
#include "npstat/stat/AbsDistribution1D.hh"

namespace npstat {
    /**
    // This class will construct a local polynomial filter out of
    // an arbitrary sampled even weight function.
    */
    class WeightTableFilter1DBuilder : public OrthoPolyFilter1DBuilder
    {
    public:
        /**
        // The constructor arguments are as follows:
        //
        //  weight    -- Tabulated weight function, in equidistant
        //               steps, in the order of increasing x.
        //               The first value corresponds to x = 0,
        //               the second to x + step, etc. The weight
        //               is assumed to be symmetric around 0.
        //
        //  weightLen -- The length of the "weight" array.
        //
        //  exclusionMask    -- Set values of "exclusionMask" != 0 if
        //                      corresponding data points have to be
        //                      excluded when weights are generated.
        //                      If no exclusions are necessary, just
        //                      leave this array as NULL.
        //
        //  exclusionMaskLen -- Length of the "exclusionMask" array. If
        //                      it is not 0 then it must coinside with
        //                      the "datalen" argument given to all future
        //                      invocations of the "makeFilter" method.
        //
        //  excludeCentralPoint -- If "true", the central point of the
        //                      weight will be set to zero. This can be
        //                      useful for certain types of cross validation
        //                      calculations.
        */
        WeightTableFilter1DBuilder(const double* weight, unsigned weightLen,
                                   const unsigned char* exclusionMask = 0,
                                   unsigned exclusionMaskLen = 0,
                                   bool excludeCentralPoint = false);

        inline virtual ~WeightTableFilter1DBuilder() {}

        inline unsigned centralWeightLength() const {return w_.size();}
        inline bool keepAllFilters() const {return !exclusionMask_.empty();}

        virtual OrthoPoly1D* makeOrthoPoly(unsigned maxDegree,
                                           unsigned binnum,
                                           unsigned datalen,
                                           unsigned* filterCenter) const;
    private:
        WeightTableFilter1DBuilder();

        std::vector<double> w_;
        mutable std::vector<double> wexcl_;
        std::vector<unsigned char> exclusionMask_;
    };

    /**
    // This class will construct a local polynomial filter out of
    // an AbsDistribution1D weight function assumed to be even.
    //
    // The weight function will be stretched near the boundaries
    // by the factor Iw_max / Iw, where Iw is the integral of the
    // weight function inside the support boundary (Iw_max is the
    // same integral calculated far away from any boundary).
    */
    class StretchingFilter1DBuilder : public OrthoPolyFilter1DBuilder
    {
    public:
        /**
        // This class will not own the AbsDistribution1D object.
        // It is the responsibility of the user of this class
        // to make sure that the AbsDistribution1D object stays
        // alive while this object is in use.
        //
        // It will be assumed that the distribution has its peak at 0.0.
        */
        StretchingFilter1DBuilder(const AbsDistribution1D* distro,
                                  double centralStepSize,
                                  const unsigned char* exclusionMask = 0,
                                  unsigned exclusionMaskLen = 0,
                                  bool excludeCentralPoint = false);

        inline virtual ~StretchingFilter1DBuilder() {}

        inline unsigned centralWeightLength() const {return maxlen_;}
        inline bool keepAllFilters() const {return !exclusionMask_.empty();}

        virtual OrthoPoly1D* makeOrthoPoly(unsigned maxDegree,
                                           unsigned binnum,
                                           unsigned datalen,
                                           unsigned* filterCenter) const;
    private:
        StretchingFilter1DBuilder();

        const AbsDistribution1D* distro_;
        double step_;
        unsigned maxlen_;
        mutable std::vector<double> w_;
        std::vector<unsigned char> exclusionMask_;
        bool excludeCentralPoint_;
    };

    /**
    // This class will construct a local polynomial filter out of
    // an AbsDistribution1D weight function assumed to be even.
    //
    // The weight function will be stretched near the boundaries
    // so that the integral of the weight function squared is
    // preserved when the weight is normalized.
    */
    class ConstSqFilter1DBuilder : public OrthoPolyFilter1DBuilder
    {
    public:
        /**
        // This class will not own the AbsDistribution1D object.
        // It is the responsibility of the user of this class
        // to make sure that the AbsDistribution1D object stays
        // alive while this object is in use.
        //
        // It will be assumed that the distribution has its peak at 0.0.
        */
        ConstSqFilter1DBuilder(const AbsDistribution1D* distro,
                               double centralStepSize,
                               const unsigned char* exclusionMask = 0,
                               unsigned exclusionMaskLen = 0,
                               bool excludeCentralPoint = false);

        inline virtual ~ConstSqFilter1DBuilder() {}

        inline unsigned centralWeightLength() const {return maxlen_;}
        inline bool keepAllFilters() const {return !exclusionMask_.empty();}

        virtual OrthoPoly1D* makeOrthoPoly(unsigned maxDegree,
                                           unsigned binnum,
                                           unsigned datalen,
                                           unsigned* filterCenter) const;
    private:
        ConstSqFilter1DBuilder();

        double integrateSquaredWeight(unsigned datalen, unsigned binnum,
                                      double h) const;
        class Fnc;
        friend class Fnc;
        class Fnc : public Functor1<double,double>
        {
        public:
            inline Fnc(const ConstSqFilter1DBuilder& ref,
                       const unsigned datalen, const unsigned binnum)
                : ref_(ref), datalen_(datalen), binnum_(binnum) {}
            inline virtual ~Fnc() {}

            inline double operator()(const double& h) const
                {return ref_.integrateSquaredWeight(datalen_, binnum_, h);}

        private:
            const ConstSqFilter1DBuilder& ref_;
            unsigned datalen_;
            unsigned binnum_;
        };

        const AbsDistribution1D* distro_;
        double step_;
        double integ_;
        unsigned maxlen_;
        mutable std::vector<double> w_;
        std::vector<unsigned char> exclusionMask_;
        bool excludeCentralPoint_;
    };

    /**
    // This class will mirror (or "fold") the kernel so that it fits
    // inside the boundaries. The kernel is assumed to be an even function.
    */
    class FoldingFilter1DBuilder : public OrthoPolyFilter1DBuilder
    {
    public:
        /**
        // This class will not own the AbsDistribution1D object.
        // It is the responsibility of the user of this class
        // to make sure that the AbsDistribution1D object stays
        // alive while this object is in use.
        //
        // It will be assumed that the distribution has its peak at 0.0.
        */
        FoldingFilter1DBuilder(const AbsDistribution1D* distro,
                               double stepSize,
                               const unsigned char* exclusionMask = 0,
                               unsigned exclusionMaskLen = 0,
                               bool excludeCentralPoint = false);

        inline virtual ~FoldingFilter1DBuilder() {}

        inline unsigned centralWeightLength() const {return maxlen_;}
        inline bool keepAllFilters() const {return !exclusionMask_.empty();}

        virtual OrthoPoly1D* makeOrthoPoly(unsigned maxDegree,
                                           unsigned binnum,
                                           unsigned datalen,
                                           unsigned* filterCenter) const;
    private:
        FoldingFilter1DBuilder();

        const AbsDistribution1D* distro_;
        double step_;
        unsigned maxlen_;
        mutable std::vector<double> w_;
        std::vector<unsigned char> exclusionMask_;
        bool excludeCentralPoint_;
    };

    /**
    // This class will both fold and stretch the kernel to preserve its
    // squared integral. The kernel is assumed to be an even function.
    */
    class FoldingSqFilter1DBuilder : public OrthoPolyFilter1DBuilder
    {
    public:
        /**
        // This class will not own the AbsDistribution1D object.
        // It is the responsibility of the user of this class
        // to make sure that the AbsDistribution1D object stays
        // alive while this object is in use.
        //
        // It will be assumed that the distribution has its peak at 0.0.
        */
        FoldingSqFilter1DBuilder(const AbsDistribution1D* distro,
                                 double stepSize,
                                 const unsigned char* exclusionMask = 0,
                                 unsigned exclusionMaskLen = 0,
                                 bool excludeCentralPoint = false);

        inline virtual ~FoldingSqFilter1DBuilder() {}

        inline unsigned centralWeightLength() const {return maxlen_;}
        inline bool keepAllFilters() const {return !exclusionMask_.empty();}

        virtual OrthoPoly1D* makeOrthoPoly(unsigned maxDegree,
                                           unsigned binnum,
                                           unsigned datalen,
                                           unsigned* filterCenter) const;
    private:
        FoldingSqFilter1DBuilder();

        double integrateSquaredWeight(unsigned datalen, unsigned binnum,
                                      double h) const;
        class Fnc;
        friend class Fnc;
        class Fnc : public Functor1<double,double>
        {
        public:
            inline Fnc(const FoldingSqFilter1DBuilder& ref,
                       const unsigned datalen, const unsigned binnum)
                : ref_(ref), datalen_(datalen), binnum_(binnum) {}
            inline virtual ~Fnc() {}

            inline double operator()(const double& h) const
                {return ref_.integrateSquaredWeight(datalen_, binnum_, h);}

        private:
            const FoldingSqFilter1DBuilder& ref_;
            unsigned datalen_;
            unsigned binnum_;
        };

        const AbsDistribution1D* distro_;
        double step_;
        double integ_;
        unsigned maxlen_;
        mutable std::vector<double> w_;
        std::vector<unsigned char> exclusionMask_;
        bool excludeCentralPoint_;
    };

    /**
    // This class will construct a filter which does not smooth the input.
    // Such a filter is sometimes useful for testing purposes.
    */
    class NonmodifyingFilter1DBuilder : public WeightTableFilter1DBuilder
    {
    public:
        NonmodifyingFilter1DBuilder();

        inline virtual ~NonmodifyingFilter1DBuilder() {}

        virtual OrthoPoly1D* makeOrthoPoly(unsigned maxDegree,
                                           unsigned binnum,
                                           unsigned datalen,
                                           unsigned* filterCenter) const;
    };
}

#endif // NPSTAT_FILTER1DBUILDERS_HH_
