#include <cmath>
#include <cfloat>
#include <utility>
#include <stdexcept>

#include "geners/GenericIO.hh"
#include "npstat/stat/VerticallyInterpolatedDistro1D1P.hh"

namespace npstat {
    VerticallyInterpolatedDistro1D1P::VerticallyInterpolatedDistro1D1P(
        const GridAxis& axis,
        const std::vector<CPP11_shared_ptr<const AbsDistribution1D> >& distros,
        const AbsDistributionTransform1D& transform)
        : axis_(axis),
          distros_(distros),
          t_(transform.clone()),
          changed_(true)
    {
        const unsigned nDistros = distros_.size();
        if (!nDistros) throw  std::invalid_argument(
            "In npstat::VerticallyInterpolatedDistro1D1P constructor: "
            "no distros to interpolate");
        if (axis_.nCoords() != nDistros) throw std::invalid_argument(
            "In npstat::VerticallyInterpolatedDistro1D1P constructor: "
            "number of provided distributions must be equal to the grid size");
        if (!t_->nParameters()) throw  std::invalid_argument(
            "In npstat::VerticallyInterpolatedDistro1D1P constructor: "
            "transform does not have any parameters");
    }

    bool VerticallyInterpolatedDistro1D1P::isEqual(
        const AbsDistribution1D& otherBase) const
    {
        const VerticallyInterpolatedDistro1D1P& r =
            static_cast<const VerticallyInterpolatedDistro1D1P&>(otherBase);
        if (*t_ != *r.t_)
            return false;
        if (axis_ != r.axis_)
            return false;
        const unsigned sz = distros_.size();
        if (sz != r.distros_.size())
            return false;
        for (unsigned i=0; i<sz; ++i)
            if (*distros_[i] != *r.distros_[i])
                return false;
        return true;
    }

    bool VerticallyInterpolatedDistro1D1P::write(std::ostream& os) const
    {
        return gs::write_obj(os, axis_) &&
               gs::write_item(os, distros_) &&
               gs::write_obj(os, *t_);
    }

    VerticallyInterpolatedDistro1D1P* VerticallyInterpolatedDistro1D1P::read(
        const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(
            gs::ClassId::makeId<VerticallyInterpolatedDistro1D1P>());
        current.ensureSameId(id);

        CPP11_auto_ptr<GridAxis> paxis = gs::read_obj<GridAxis>(in);
        std::vector<CPP11_shared_ptr<const AbsDistribution1D> > distros;
        gs::restore_item(in, &distros);
        CPP11_auto_ptr<AbsDistributionTransform1D> p =
            gs::read_obj<AbsDistributionTransform1D>(in);
        return new VerticallyInterpolatedDistro1D1P(*paxis, distros, *p);
    }

    VerticallyInterpolatedDistro1D1P::VerticallyInterpolatedDistro1D1P(
        const VerticallyInterpolatedDistro1D1P& r)
        : axis_(r.axis_),
          distros_(r.distros_),
          t_(r.t_->clone()),
          changed_(true)
    {
    }

    VerticallyInterpolatedDistro1D1P&
    VerticallyInterpolatedDistro1D1P::operator=(
        const VerticallyInterpolatedDistro1D1P& r)
    {
        if (this != &r)
        {
            AbsDistribution1D::operator=(r);
            axis_ = r.axis_;
            distros_ = r.distros_;
            delete t_; t_ = 0;
            t_ = r.t_->clone();
            changed_ = true;
        }
        return *this;
    }

    void VerticallyInterpolatedDistro1D1P::calcWeights() const
    {
        const double p = t_->getParameter(0U);
        const std::pair<unsigned,double>& intv = axis_.getInterval(p);
        cell_ = intv.first;
        wleft_ = intv.second;
        wright_ = 1.0 - wleft_;
        changed_ = false;
    }

    double VerticallyInterpolatedDistro1D1P::density(const double x) const
    {
        if (changed_) calcWeights();
        double deriv;
        const double y = t_->transformForward(x, &deriv);
        double dsum = 0.0;
        if (wleft_ > 0.0)
            dsum += wleft_*distros_[cell_]->density(y);
        if (wright_ > 0.0)
            dsum += wright_*distros_[cell_+1U]->density(y);
        return dsum*fabs(deriv);
    }

    double VerticallyInterpolatedDistro1D1P::cdf_in_y(const double y) const
    {
        double dsum = 0.0;
        if (wleft_ > 0.0)
            dsum += wleft_*distros_[cell_]->cdf(y);
        if (wright_ > 0.0)
            dsum += wright_*distros_[cell_+1U]->cdf(y);
        return dsum;
    }

    double VerticallyInterpolatedDistro1D1P::cdf(const double x) const
    {
        if (changed_) calcWeights();
        const double c = cdf_in_y(t_->transformForward(x, 0));
        return t_->isIncreasing() ? c : 1.0 - c;
    }

    double VerticallyInterpolatedDistro1D1P::exceedance(const double x) const
    {
        return 1.0 - cdf(x);
    }

    double VerticallyInterpolatedDistro1D1P::quantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::VerticallyInterpolatedDistro1D1P::quantile: "
            "cdf argument outside of [0, 1] interval");
        if (changed_) calcWeights();
        const double rtmp = t_->isIncreasing() ? r1 : 1.0 - r1;
        const double q = quantile_in_y(rtmp);
        return t_->transformBack(q);
    }

    double VerticallyInterpolatedDistro1D1P::quantile_in_y(const double r1) const
    {
        double qmax = distros_[cell_]->quantile(r1);
        double qmin = distros_[cell_+1U]->quantile(r1);
        if (!(qmin <= qmax)) std::swap(qmin, qmax);
        if (qmax == qmin)
            return qmin;
        if (r1 == 1.0)
            return qmax;
        if (r1 == 0.0)
            return qmin;
        const double fmin = cdf_in_y(qmin);
        const double fmax = cdf_in_y(qmax);
        if (!(fmin < r1 && r1 < fmax))
            throw std::runtime_error(
                "In npstat::VerticallyInterpolatedDistro1D1P::quantile_in_y: "
                "algorithm precondition error");
        for (unsigned i=0; i<1000; ++i)
        {
            const double x = (qmin + qmax)/2.0;
            if (fabs(qmax - qmin)/std::max(fabs(qmin), fabs(qmax)) <
                2.0*DBL_EPSILON)
                return x;
            const double fval = cdf_in_y(x);
            if (fval == r1)
                return x;
            else if (fval > r1)
                qmax = x;
            else
                qmin = x;
            if (qmax == qmin)
                return qmin;
        }
        return (qmin + qmax)/2.0;
    }
}
